/**
 * Base types and functions for the runtime implementation.
 *
 * @author Igor Konnov, Gabriela Moreira
 *
 * @module
 */
import { Maybe } from '@sweet-monads/maybe';
import { ErrorCode, QuintError } from '../../quintError';
import { Computable, EvaluationResult, Register } from '../runtime';
import { ExecutionListener } from '../trace';
import { Trace } from './trace';
import { Either } from '@sweet-monads/either';
import { RuntimeValue } from './runtimeValue';
export declare const specialNames: string[];
/**
 * Returns a Map containing the built-in Computable objects for the Quint language.
 * These include the callable objects for Bool, Int, and Nat.
 *
 * @returns a Map containing the built-in Computable objects.
 */
export declare function builtinContext(): Map<string, Computable>;
/**
 * Represents the state of evaluation of Quint code.
 * All the fields are mutated by CompilerVisitor, either directly, or via calls.
 */
export interface EvaluationState {
    context: Map<string, Computable>;
    vars: Register[];
    nextVars: Register[];
    trace: Trace;
    errorTracker: CompilerErrorTracker;
    listener: ExecutionListener;
}
/**
 * Creates a new EvaluationState object.
 *
 * @returns a new EvaluationState object.
 */
export declare class CompilerErrorTracker {
    compileErrors: QuintError[];
    runtimeErrors: QuintError[];
    addCompileError(reference: bigint, code: ErrorCode, message: string): void;
    addRuntimeError(reference: bigint | undefined, error: QuintError): void;
}
/**
 * Creates a new EvaluationState object with the initial state of the evaluation.
 *
 * @returns a new EvaluationState object
 */
export declare function newEvaluationState(listener: ExecutionListener): EvaluationState;
export declare function toMaybe<T>(r: Either<QuintError, T>): Maybe<T>;
export declare function mkConstComputable(value: RuntimeValue): {
    eval: () => Either<never, RuntimeValue>;
};
export declare function mkFunComputable(fun: () => EvaluationResult): {
    eval: () => EvaluationResult;
};
