"use strict";
/* ----------------------------------------------------------------------------------
 * Copyright 2022-2024 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 * --------------------------------------------------------------------------------- */
Object.defineProperty(exports, "__esModule", { value: true });
exports.mkFunComputable = exports.mkConstComputable = exports.toMaybe = exports.newEvaluationState = exports.CompilerErrorTracker = exports.builtinContext = exports.specialNames = void 0;
/**
 * Base types and functions for the runtime implementation.
 *
 * @author Igor Konnov, Gabriela Moreira
 *
 * @module
 */
const maybe_1 = require("@sweet-monads/maybe");
const runtime_1 = require("../runtime");
const trace_1 = require("./trace");
const either_1 = require("@sweet-monads/either");
const runtimeValue_1 = require("./runtimeValue");
// Internal names in the compiler, which have special treatment.
// For some reason, if we replace 'q::input' with inputDefName, everything breaks.
// What kind of JS magic is that?
exports.specialNames = ['q::input', 'q::runResult', 'q::nruns', 'q::nsteps', 'q::init', 'q::next', 'q::inv'];
/**
 * Returns a Map containing the built-in Computable objects for the Quint language.
 * These include the callable objects for Bool, Int, and Nat.
 *
 * @returns a Map containing the built-in Computable objects.
 */
function builtinContext() {
    return new Map([
        [(0, runtime_1.kindName)('callable', 'Bool'), (0, runtime_1.mkCallable)([], mkConstComputable(runtimeValue_1.rv.mkSet([runtimeValue_1.rv.mkBool(false), runtimeValue_1.rv.mkBool(true)])))],
        [(0, runtime_1.kindName)('callable', 'Int'), (0, runtime_1.mkCallable)([], mkConstComputable(runtimeValue_1.rv.mkInfSet('Int')))],
        [(0, runtime_1.kindName)('callable', 'Nat'), (0, runtime_1.mkCallable)([], mkConstComputable(runtimeValue_1.rv.mkInfSet('Nat')))],
    ]);
}
exports.builtinContext = builtinContext;
/**
 * Creates a new EvaluationState object.
 *
 * @returns a new EvaluationState object.
 */
class CompilerErrorTracker {
    constructor() {
        // messages that are produced during compilation
        this.compileErrors = [];
        // messages that get populated as the compiled code is executed
        this.runtimeErrors = [];
    }
    addCompileError(reference, code, message) {
        this.compileErrors.push({ code, message, reference });
    }
    addRuntimeError(reference, error) {
        this.runtimeErrors.push({ ...error, reference });
    }
}
exports.CompilerErrorTracker = CompilerErrorTracker;
/**
 * Creates a new EvaluationState object with the initial state of the evaluation.
 *
 * @returns a new EvaluationState object
 */
function newEvaluationState(listener) {
    const state = {
        context: builtinContext(),
        vars: [],
        nextVars: [],
        trace: new trace_1.Trace(),
        errorTracker: new CompilerErrorTracker(),
        listener: listener,
    };
    return state;
}
exports.newEvaluationState = newEvaluationState;
function toMaybe(r) {
    if (r.isRight()) {
        return (0, maybe_1.just)(r.value);
    }
    else {
        return (0, maybe_1.none)();
    }
}
exports.toMaybe = toMaybe;
// make a `Computable` that always returns a given runtime value
function mkConstComputable(value) {
    return {
        eval: () => {
            return (0, either_1.right)(value);
        },
    };
}
exports.mkConstComputable = mkConstComputable;
// make a `Computable` that always returns a given runtime value
function mkFunComputable(fun) {
    return {
        eval: () => {
            return fun();
        },
    };
}
exports.mkFunComputable = mkFunComputable;
//# sourceMappingURL=base.js.map