"use strict";
/* ----------------------------------------------------------------------------------
 * Copyright 2022-2024 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 * --------------------------------------------------------------------------------- */
Object.defineProperty(exports, "__esModule", { value: true });
exports.mapLambdaThenReduce = exports.applyBoolOp = exports.sliceList = exports.updateList = exports.getListElem = exports.applyFold = exports.applyFun = void 0;
/**
 * Stateless implementations for evaluating more complex operators, and helpers
 * to operator evaluation.
 *
 * @author Igor Konnov, Gabriela Moreira
 *
 * @module
 */
const either_1 = require("@sweet-monads/either");
const runtimeValue_1 = require("./runtimeValue");
const assert_1 = require("assert");
const quintError_1 = require("../../quintError");
// pop nargs computable values, pass them the 'fun' function, and
// push the combined computable value on the stack
function applyFun(sourceId, fun, args) {
    // produce the new computable value
    return {
        eval: () => {
            try {
                // compute the values of the arguments at this point
                const values = args.map(a => a.eval());
                // if they are all defined, apply the function 'fun' to the arguments
                return (0, either_1.mergeInMany)(values)
                    .mapLeft((errors) => ({ code: 'QNT501', message: errors.map(quintError_1.quintErrorToString).join('\n') }))
                    .chain(vs => fun(...vs.map(v => v)));
            }
            catch (error) {
                const msg = error instanceof Error ? error.message : 'unknown error';
                return (0, either_1.left)({ code: 'QNT501', message: msg, reference: sourceId });
            }
        },
    };
}
exports.applyFun = applyFun;
function applyFold(order, args) {
    const [iterableComp, initialComp, callableComp] = args;
    return {
        eval: () => {
            const iterableResult = iterableComp.eval();
            const initialResult = initialComp.eval();
            const callable = callableComp;
            (0, assert_1.strict)(callable.nparams === 2);
            if (iterableResult.isLeft()) {
                return iterableResult;
            }
            const iterable = Array.from(iterableResult.value);
            const reducer = (acc, val) => {
                if (order === 'fwd') {
                    return callable.eval([acc, (0, either_1.right)(val)]);
                }
                else {
                    return callable.eval([(0, either_1.right)(val), acc]);
                }
            };
            if (order === 'fwd') {
                return iterable.reduce(reducer, initialResult);
            }
            else {
                return iterable.reduceRight(reducer, initialResult);
            }
        },
    };
}
exports.applyFold = applyFold;
// Access a list via an index
function getListElem(list, idx) {
    if (idx >= 0n && idx < list.size) {
        const elem = list.get(Number(idx));
        if (elem) {
            return (0, either_1.right)(elem);
        }
    }
    return (0, either_1.left)({ code: 'QNT510', message: `Out of bounds, nth(${idx})` });
}
exports.getListElem = getListElem;
// Update a list via an index
function updateList(list, idx, value) {
    if (idx >= 0n && idx < list.size) {
        return (0, either_1.right)(runtimeValue_1.rv.mkList(list.set(Number(idx), value)));
    }
    else {
        return (0, either_1.left)({ code: 'QNT510', message: `Out of bounds, replaceAt(..., ${idx}, ...)` });
    }
}
exports.updateList = updateList;
// slice a list
function sliceList(list, start, end) {
    return (0, either_1.right)(runtimeValue_1.rv.mkList(list.slice(start, end)));
}
exports.sliceList = sliceList;
// translate one of the Boolean operators with short-circuiting:
//  - or  { A, ..., C }
//  - and { A, ..., C }
//  - A implies B
function applyBoolOp(defaultValue, shortCircuit, args) {
    return {
        eval: () => {
            let result = (0, either_1.right)(defaultValue);
            // Evaluate arguments iteratively.
            // Stop as soon as shortCircuit tells us to stop.
            let no = 0;
            for (const arg of args) {
                // either the argument is evaluated to a Boolean, or fails
                result = arg.eval();
                if (result.isLeft()) {
                    return result;
                }
                // if shortCircuit returns a value, return the value immediately
                const b = shortCircuit(no, result.value.toBool());
                if (b) {
                    return (0, either_1.right)(b);
                }
                no += 1;
            }
            return result;
        },
    };
}
exports.applyBoolOp = applyBoolOp;
/**
 * A generalized application of a one-argument Callable to a set-like
 * runtime value, as required by `exists`, `forall`, `map`, and `filter`.
 *
 * This method expects `compStack` to look like follows:
 *
 *  - `(top)` translated lambda, as `Callable`.
 *
 *  - `(top - 1)`: a set-like value to iterate over, as `Computable`.
 *
 * The method evaluates the Callable for each element of the iterable value
 * and either produces `none`, if evaluation failed for one of the elements,
 * or it applies `mapResultAndElems` to the pairs that consists of the Callable
 * result and the original element of the iterable value.
 * The final result is stored on the stack.
 */
function mapLambdaThenReduce(sourceId, reduceFunction, args) {
    const [setComp, callableComp] = args;
    const callable = callableComp;
    // apply the lambda to a single element of the set
    const evaluateElem = (elem) => {
        // evaluate the predicate against the actual arguments
        const result = callable.eval([(0, either_1.right)(elem)]);
        return result.map(result => [result, elem]);
    };
    return applyFun(sourceId, (iterable) => {
        const lambdaApplicationResults = (0, either_1.mergeInMany)(Array.from(iterable).map(value => evaluateElem(value)));
        return lambdaApplicationResults
            .mapLeft((errors) => {
            return { code: 'QNT501', message: errors.map(quintError_1.quintErrorToString).join('\n') };
        })
            .map(array => reduceFunction(array));
    }, [setComp]);
}
exports.mapLambdaThenReduce = mapLambdaThenReduce;
//# sourceMappingURL=operatorEvaluator.js.map