"use strict";
/*
 * Runtime environment for Quint. This runtime environment is designed for the
 * following purposes:
 *
 *  - Give the language user fast feedback via interpretation of expressions.
 *  - Let the user simulate their specification similar to property-based testing.
 *
 * Igor Konnov, 2022-2023
 *
 * Copyright 2022-2023 Informal Systems
 * Licensed under the Apache License, Version 2.0.
 * See LICENSE in the project root for license information.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.fail = exports.mkCallable = exports.mkRegister = exports.kindName = void 0;
const runtimeValue_1 = require("./impl/runtimeValue");
const either_1 = require("@sweet-monads/either");
const base_1 = require("./impl/base");
/**
 * Create a key that encodes its name and kind. This is only useful for
 * storing registers in a map, as JS objects are hard to use as keys.
 * In a good language, we would use (kind, name), but not in JS.
 */
const kindName = (kind, name) => {
    return `${kind}:${name}`;
};
exports.kindName = kindName;
/**
 * Create an object that implements Register.
 */
function mkRegister(kind, registerName, initValue, errorForMissing) {
    const reg = {
        name: registerName,
        kind,
        registerValue: initValue,
        // first, define a fruitless eval, as we cannot refer to registerValue yet
        eval: () => {
            return (0, either_1.left)(errorForMissing);
        },
    };
    // override `eval`, as we can use `reg` now
    reg.eval = () => {
        // computing a register just evaluates to the contents that it stores
        if (reg.registerValue.isNone()) {
            return (0, either_1.left)(errorForMissing);
        }
        return (0, either_1.right)(reg.registerValue.value);
    };
    return reg;
}
exports.mkRegister = mkRegister;
/**
 * Create an object that implements Callable.
 */
function mkCallable(registers, body) {
    const callable = {
        nparams: registers.length,
        eval: () => body.eval(),
    };
    callable.eval = args => {
        if (registers.length === 0) {
            // simply evaluate the body, no parameters are needed
            return body.eval();
        }
        if (args && args.length >= registers.length) {
            // All parameters are passed via `args`. Store them in the registers.
            registers.forEach((r, i) => (r.registerValue = (0, base_1.toMaybe)(args[i])));
            // Evaluate the body under for the registers set to `args`.
            return body.eval();
        }
        else {
            // The lambda is evaluated without giving the arguments.
            // All we can do is to return this lambda as a runtime value.
            return (0, either_1.right)(runtimeValue_1.rv.mkLambda(registers.length, callable));
        }
    };
    return callable;
}
exports.mkCallable = mkCallable;
/**
 * An implementation of Computable that always fails.
 */
exports.fail = {
    eval: () => {
        return (0, either_1.left)({ code: 'QNT501', message: 'Failed to evaluate something :/' });
    },
};
//# sourceMappingURL=runtime.js.map